@@ -1,11 +1,15 @@
 package Test::MockObject;
 
 use strict;
+use warnings;
 
 use vars qw( $VERSION $AUTOLOAD );
-$VERSION = '0.20';
+$VERSION = '1.07';
+
+use Scalar::Util qw( blessed refaddr reftype weaken );
+use UNIVERSAL::isa;
+use UNIVERSAL::can;
 
-use Scalar::Util qw( reftype blessed );
 use Test::Builder;
 
 my $Test = Test::Builder->new();
@@ -30,16 +34,6 @@ sub mock
 	$self;
 }
 
-# deprecated and complicated as of 0.07
-sub add
-{
-	my $self = shift;
-	my $subs = _subs( $self );
-	return $subs->{add}->( $self, @_ )
-		 if (exists $subs->{add} and !( UNIVERSAL::isa( $_[1], 'CODE' )));
-	$self->mock( @_ );
-}
-
 sub set_isa 
 {
     my ($self, @supers) = @_;
@@ -98,32 +92,42 @@ sub set_bound
 		SCALAR => sub { $$ref },
 		ARRAY  => sub { @$ref },
 		HASH   => sub { %$ref },
-
 	);
 
 	return unless exists $bindings{reftype( $ref )};
 	$self->mock( $name,  $bindings{reftype( $ref )} );
 }
 
+# hack around debugging mode being too smart for my sub names
+my $old_p;
 BEGIN
 {
-	no strict 'refs';
+	$old_p  = $^P;
+	$^P    &= ~0x200;
+}
 
-	for my $universal (
-		{ sub => \&_subs, name => 'can', parent => \&UNIVERSAL::can },
-		{ sub => \&_isas, name => 'isa', parent => \&UNIVERSAL::isa },
-	)
+BEGIN
+{
+	for my $universal
+	( { sub => \&_subs, name => 'can' }, { sub => \&_isas, name => 'isa' } )
 	{
-		*{ $universal->{name} } = sub
+		my $sub = sub
 		{
 			my ($self, $sub) = @_; 
+			local *__ANON__  = $universal->{name};
 
 			# mockmethods are special cases, class methods are handled directly
 			my $lookup = $universal->{sub}->( $self );
 			return $lookup->{$sub} if blessed $self and exists $lookup->{$sub};
-			return $universal->{parent}->( @_);
+			my $parent = 'SUPER::' . $universal->{name};
+			return $self->$parent( $sub );
 		};
+
+		no strict 'refs';
+		*{ $universal->{name} } = $sub;
 	}
+
+	$^P = $old_p;
 }
 
 sub remove
@@ -136,8 +140,9 @@ sub remove
 sub called
 {
 	my ($self, $sub) = @_;
-	
-	for my $called (reverse @{ _calls( $self ) }) {
+
+	for my $called (reverse @{ _calls( $self ) })
+	{
 		return 1 if $called->[0] eq $sub;
 	}
 
@@ -225,15 +230,23 @@ sub dispatch_mocked_method
 		require Carp;
 		Carp::carp("Un-mocked method '$sub()' called");
 	}
+
 	return;
 }
 
 sub log_call
 {
-	my ($self, $sub) = splice( @_, 0, 2 );
-
+	my ($self, $sub, @call_args) = @_;
 	return unless _logs( $self, $sub );
-	push @{ _calls( $self ) }, [ $sub, [ @_ ] ];
+
+	# prevent circular references with weaken
+	for my $arg ( @call_args )
+	{
+		next unless ref $arg;
+		weaken( $arg ) if refaddr( $arg ) eq refaddr( $self );
+	}		
+
+	push @{ _calls( $self ) }, [ $sub, \@call_args ];
 }
 
 sub called_ok
@@ -272,13 +285,23 @@ sub called_args_pos_is
 sub fake_module
 {
 	my ($class, $modname, %subs) = @_;
+
+	if ($class->check_class_loaded( $modname ) and ! keys %subs)
+	{
+		require Carp;
+		Carp::croak( "No mocked subs for loaded module '$modname'" );
+	}
+
 	$modname =~ s!::!/!g;
 	$INC{ $modname . '.pm' } = 1;
 
-	local $SIG{__WARN__} = sub { warn $_[0] unless $_[0] =~ /redefined/ };
-	no strict 'refs';
-	${ $modname . '::' }{VERSION} ||= -1;
-	
+	my $warn = $SIG{__WARN__};
+	local $SIG{__WARN__} = sub { $warn->( $_[0] ) unless $_[0] =~ /redefined/ };
+	{
+		no strict 'refs';
+		${ $modname . '::' }{VERSION} ||= -1;
+	}
+
 	for my $sub (keys %subs)
 	{
 		my $type = reftype( $subs{ $sub } ) || '';
@@ -288,23 +311,67 @@ sub fake_module
 			Carp::carp("'$sub' is not a code reference" );
 			next;
 		}
+		no strict 'refs';
 		*{ $_[1] . '::' . $sub } = $subs{ $sub };
 	}
 }
 
+sub check_class_loaded
+{
+	my ($self, $class, $load_flag) = @_;
+
+	(my $path    = $class) =~ s{::}{/}g;
+	return 1 if exists $INC{ $path . '.pm' };
+
+	my $symtable = \%main::;
+	my $found    = 1;
+
+	for my $symbol ( split( '::', $class ))
+	{
+		unless (exists $symtable->{ $symbol . '::' })
+		{
+			$found = 0;
+			last;
+		}
+
+		$symtable = $symtable->{ $symbol . '::' };
+	}
+
+	return $found;
+}
+
 sub fake_new
 {
 	my ($self, $class) = @_;
 	$self->fake_module( $class, new => sub { $self } );
 }
 
+sub DESTROY
+{
+	my $self = shift;
+	$self->_clear_calls();
+	$self->_clear_subs();
+	$self->_clear_logs();
+	$self->_clear_isas();
+}
+
+sub _get_key
+{
+	my $invocant = shift;
+	return blessed( $invocant ) ? refaddr( $invocant ) : $invocant;
+}
+
 {
 	my %calls;
 
 	sub _calls
 	{
-		my $key = shift;
-		$calls{ $key } ||= [];
+		$calls{ _get_key( shift ) } ||= [];
+	}
+
+	sub _clear_calls
+	{
+		delete $calls{ _get_key( shift ) };
 	}
 }
 
@@ -313,8 +380,12 @@ sub fake_new
 
 	sub _subs
 	{
-		my $key = shift;
-		$subs{ $key } ||= {};
+		$subs{ _get_key( shift ) } ||= {};
+	}
+
+	sub _clear_subs
+	{
+		delete $subs{ _get_key( shift ) };
 	}
 }
 
@@ -323,7 +394,8 @@ sub fake_new
 
 	sub _set_log
 	{
-		my ($key, $name, $log) = @_;
+		my $key          = _get_key( shift );
+		my ($name, $log) = @_;
 
 		$logs{$key} ||= {};
 
@@ -339,9 +411,15 @@ sub fake_new
 
 	sub _logs
 	{
-		my ($key, $name) = @_;
+		my $key    = _get_key( shift );
+		my ($name) = @_;
 		return exists $logs{$key}{$name};
 	}
+
+	sub _clear_logs
+	{
+		delete $logs{ _get_key( shift ) };
+	}
 }
 
 {
@@ -349,8 +427,12 @@ sub fake_new
     
 	sub _isas
 	{
-		my $key = shift;
-		$isas{ $key } ||= {};
+		$isas{ _get_key( shift ) } ||= {};
+	}
+
+	sub _clear_isas
+	{
+		delete $isas{ _get_key( shift ) };
 	}
 }
 
@@ -490,6 +572,12 @@ C<import()>:
 	is( $import, 'Regexp::Esperanto',
 		'Regexp::Esperanto should use() Regexp::English' );
 
+If you use C<fake_module()> to mock a module that already exists in memory --
+one you've loaded elsewhere perhaps, but do not pass any subroutines to mock,
+this method will throw an exception.  This is because if you call the
+constructor later on, you probably won't get a mock object back and you'll be
+confused.
+
 =item * C<fake_new(I<module name>)>
 
 B<Note:> see L<Test::MockObject::Extends> for a better alternative to this
@@ -540,7 +628,7 @@ Adds a method bound to a variable.  Pass in a reference to a variable in your
 test.  When you change the variable, the return value of the new method will
 change as well.  This is often handier than replacing mock methods.
 
-=item * C<set_isas( I<name1>, I<name2>, ... I<namen> )>
+=item * C<set_isa( I<name1>, I<name2>, ... I<namen> )>
 
 Adds an apparent parent to the module, so that calling C<isa()> on the mock
 will return true appropriately.  Sometimes you really need this.
@@ -555,6 +643,16 @@ Removes a named method.
 
 =over 4
 
+=item * C<can( $method_name )>
+
+Returns a subroutine reference if this particular mocked object can handle the
+named method, false otherwise.
+
+=item * C<isa( $class_name )>
+
+Returns true if the invocant object mocks a particular class.  You must have
+used C<set_isa()> first.
+
 =item * C<called(I<name>)>
 
 Checks to see if something has called a named method on the object.  This
@@ -695,6 +793,11 @@ Joins together all of the arguments to a method at the appropriate position and
 matches against a specified string.  A generically bland test name is provided
 by default.  You can probably do much better.
 
+=item * C<check_class_loaded( $class_name )>
+
+Attempts to determine whether you have a class of the given name loaded and
+compiled.  Returns true or false.
+
 =back
 
 =head3 Logging
@@ -711,17 +814,27 @@ will set mock both C<foo()> and C<bar()>, causing both to return true.
 However, the object will log only calls to C<bar()>, not C<foo()>.  To log
 C<foo()> again, merely mock it again without the leading C<->:
 
-	$mock->set_true( '-foo' );
+	$mock->set_true( 'foo' );
 
 C<$mock> will log all subsequent calls to C<foo()> again.
 
 =head3 Subclassing
 
-If you want to subclass this module to override any behavior, see
-C<dispatch_mocked_method>.  This module uses this method to determine how to
-call a method not available in this class.  It also controls logging.  You may
-or may not find it useful, but I certainly take advantage of it for
-Test::MockObject::Extends.
+There are two methods provided for subclassing:
+
+=over 4
+
+=item * C<dispatch_mocked_method( $method_name, @_ )>
+
+This method determines how to call a method (named as C<$method_name>) not
+available in this class.  It also controls logging.  You may or may not find it
+useful, but I certainly take advantage of it for Test::MockObject::Extends.
+
+=item * C<log_call( $method_name, @_ )>
+
+This method tracks the call of the named method and its arguments.
+
+=back
 
 =head1 TODO
 
@@ -729,17 +842,8 @@ Test::MockObject::Extends.
 
 =item * Add a factory method to avoid namespace collisions (soon)
 
-=item * Handle C<isa()> -- done in Test::MockObject::Extends
-
-=item * Make C<fake_module()> and C<fake_new()> undoable -- done in
-Test::MockObject::Extends
-
-=item * Allow mocking but not logging certain methods (Piers' suggestion, done)
-
 =item * Add more useful methods (catch C<import()>?)
 
-=item * Move C<fake_module()> and C<fake_new()> into a Test::MockModule
-
 =back
 
 =head1 AUTHOR
@@ -757,6 +861,10 @@ Test::MockObject::Extends.
 
 Stevan Little provided the impetus and code for C<set_isa()>.
 
+Nicholas Clark found a documentation error.
+
+Mutant suggested a potential problem with fake_module().
+
 =head1 SEE ALSO
 
 L<perl>, L<Test::Tutorial>, L<Test::More>,
@@ -766,10 +874,10 @@ L<http:E<sol>E<sol>www.perl.comE<sol>pubE<sol>aE<sol>2002E<sol>07E<sol>10E<sol>t
 
 =head1 COPYRIGHT
 
-Copyright 2002 - 2004 by chromatic E<lt>chromatic at wgz dot orgE<gt>.
+Copyright (c) 2002 - 2006 by chromatic E<lt>chromatic at wgz dot orgE<gt>.
 
-This program is free software; you can redistribute it and/or modify it under
-the same terms as Perl itself.
+This program is free software; you can use, modify, and redistribute it under
+the same terms as Perl 5.8.x itself.
 
 See http://www.perl.com/perl/misc/Artistic.html
 
